@@ -74,12 +74,6 @@ sub cgi_env {
     return %env;
 }
 
-sub cgi_var {
-    my ($self,$key) = @_;
-    my $value = exists($self->{env}->{$key}) ? $self->{env}->{$key} : $self->{r}->cgi_var($key);
-    return $value;
-}
-
 sub dir_config {
     my ($self,$key) = @_;
     return $self->{dir_config}->{$key};
@@ -114,7 +108,9 @@ sub send_http_header {
 
 package HTML::Mason::Preview;
 use HTML::Mason::ApacheHandler;
+use HTML::Mason::Config;
 use HTML::Mason::Tools qw(date_delta_to_secs html_escape);
+use MLDBM ($HTML::Mason::Config{mldbm_use_db}, $HTML::Mason::Config{mldbm_serializer});
 use IO::File qw(!/^SEEK/);
 use POSIX;
 
@@ -137,14 +133,14 @@ sub open_preview_settings
 
 sub handle_preview_request
 {
-    my $result = eval('handle_preview_request_1(@_)');
+    my $result = eval { handle_preview_request_1(@_) };
     my $err = $@;
     if ($err) {
 	my ($r) = @_;
 	$r->content_type('text/html');
 	$r->header_out(Pragma=>'no-cache');
 	$r->send_http_header();
-	$r->print("<h2>Previewer Error</h2>\n<pre>\n$err\n</pre>\n");
+	print("<h2>Previewer Error</h2>\n<pre>\n$err\n</pre>\n");
 	return;
     }
     return $result;
@@ -154,7 +150,7 @@ sub handle_preview_request_1
 {
     my ($r, $ah, %options) = @_;
     my (%simhdr);
-    my $createObjectLink = $options{create_object_link};
+    my $cmHome = $options{cm_home};
     
     #
     # Determine user name and port.
@@ -272,7 +268,7 @@ sub handle_preview_request_1
     }
 
     if ($conf->{output_type} eq 'HTML') {
-        $ah->handle_request($pr);
+        return $ah->handle_request($pr);
     } elsif ($conf->{output_type} eq 'Debug') {
 
 	#
@@ -292,28 +288,24 @@ sub handle_preview_request_1
 	my ($content,$i,@compEvents,$event,$key,$value,$trace);
 	my $eventnum = 0;
 	my $startCompHook = sub {
-	    my ($self) = @_;
+	    my ($req) = @_;
 	    $content .= "\cAEVENT$eventnum\cA";
-	    my $path = $self->locals->{sourceFile};
-	    my $root = $self->comp_root;
-	    $path =~ s@^$root@@;
-	    $compEvents[$eventnum++] = {type=>'startComp',path=>$path};
+	    my $path = $req->comp->title;
+	    $compEvents[$eventnum++] = {type=>'startComp',path=>$path,comp=>($req->comp)};
 	};
 	my $endCompHook = sub {
-	    my ($self) = @_;
+	    my ($req) = @_;
 	    $content .= "\cAEVENT$eventnum\cA";
-	    my $path = $self->locals->{sourceFile};
-	    my $root = $self->comp_root;
-	    $path =~ s@^$root@@;
+	    my $path = $req->comp->title;
 	    $compEvents[$eventnum++] = {type=>'endComp',path=>$path};
 	};
 	my $startFileHook = sub {
-	    my ($self,$file) = @_;
+	    my ($req,$file) = @_;
 	    $content .= "\cAEVENT$eventnum\cA";
 	    $compEvents[$eventnum++] = {type=>'startFile',path=>$file,start=>length($content)};
 	};
 	my $endFileHook = sub {
-	    my ($self,$file) = @_;
+	    my ($req,$file) = @_;
 	    my $end = length($content);
 	    $content .= "\cAEVENT$eventnum\cA";
 	    $compEvents[$eventnum++] = {type=>'endFile',path=>$file,end=>$end};
@@ -328,7 +320,9 @@ sub handle_preview_request_1
 	    my ($trace,$obj);
 	    foreach $obj (@objects) {
 		next if !defined($obj);
-		$trace .= sprintf('%s<a href="%s">%s</a>  %s%s',(' ' x (4-length($obj->{label}))),$obj->{srclink},$obj->{label},$obj->{display});
+		$trace .= sprintf('%s<a href="%s">%s</a>  %s%s',
+				  (' ' x (4-length($obj->{label}))),
+				  $obj->{srclink},$obj->{label},$obj->{display},$obj->{repeat}>1 ? " (repeats ".$obj->{repeat}." times)": "");
 		$trace .= "\n";
 	    }
 	    return $trace;
@@ -343,7 +337,7 @@ sub handle_preview_request_1
 	$interp->add_hook(name=>'preview',type=>'end_file',code=>$endFileHook);
 	$interp->out_method(sub { $content .= $_[0] });
 	$ah->output_mode(undef);
-        $ah->handle_request($pr);
+        my $statuscode = $ah->handle_request($pr);
 
 	#
 	# Validate file events. For each startFile event,
@@ -394,7 +388,7 @@ sub handle_preview_request_1
 		}
 	    }
 	}
-	
+
 	#
 	# Analyze event information. Create a list of objects
 	# (distinct component and mc_file invocations)
@@ -403,6 +397,7 @@ sub handle_preview_request_1
 	my (@objects,@stack);
 	my $objcount = 0;
 	my $fileroot = $interp->static_file_root();
+	my $comproot = $interp->comp_root();
 	foreach $event (@compEvents) {
 	    next if !defined($event);
 	    my ($type,$path) = ($event->{type},$event->{path});
@@ -410,7 +405,6 @@ sub handle_preview_request_1
 		my ($objtext,$objsrclink,$objlabel,$objtype,$objdisplay);
 		$objcount++;
 		if ($type eq 'startComp') {
-		    my $compId = $event->{compId};
 		    $objtext = "$path";
 		    $objlabel = "$objcount";
 		    $objsrclink = "#object$objcount";
@@ -424,8 +418,41 @@ sub handle_preview_request_1
 		    $objsrclink = "#object$objcount";
 		    $objtype = "file";
 		}
-		$objdisplay = ($createObjectLink) ? $createObjectLink->($objtype,$path,scalar(@stack),$objtext) : (('  ' x scalar(@stack)) . $objtext);
-		$objects[$objcount] = {count=>$objcount,text=>$objtext,display=>$objdisplay,label=>$objlabel,color=>'003399',srclink=>$objsrclink,depth=>scalar(@stack)};
+		my $lastobj = $objects[$objcount-1];
+
+		# Look for repeated entries (equal text and stack
+		# depth) and combine into one line
+		if ($objtext eq $lastobj->{text} && $objtype eq $lastobj->{type} && scalar(@stack) == $lastobj->{depth}) {
+		    $objcount--;
+		    $objects[$objcount]->{repeat}++;
+		} else {
+		    my $objdisplay = '';
+
+		    # If Content Management home specified, and we
+		    # have a file based component or a static file
+		    # inside the file root, create content management
+		    # view/edit links.
+		    if ($cmHome &&
+			(($objtype eq 'comp' and $event->{comp}->is_file_based) or
+			 ($objtype eq 'file' and $path =~ m{^$fileroot}))) {
+			my $branch;
+			if ($objtype eq 'comp') {
+			    $path = $comproot.$path;
+			    $branch = 'Components';
+			} else {
+			    $path =~ s{^$fileroot} {};
+			    $objtext =~ s{^$fileroot} {};
+			    $branch = 'Content';
+			}
+			my $spacer = ('  ' x (scalar(@stack)+1));
+			$cmHome =~ s/\/$//;
+			$objdisplay = (qq{<a href="$cmHome/textView?branch=$branch&path=$path">info</a> <a href="$cmHome/editComp?branch=$branch&path=$path">edit</a>}.$spacer.$objtext);
+		    } else {
+			$objdisplay = (('  ' x scalar(@stack)) . $objtext);
+		    }
+		    # Store object info. Some of this is used for repeat comparison.
+		    $objects[$objcount] = {count=>$objcount,type=>$objtype,display=>$objdisplay,label=>$objlabel,text=>$objtext,color=>'003399',srclink=>$objsrclink,depth=>scalar(@stack),repeat=>1};
+		}
 		push(@stack,$objcount);
 	    } elsif ($type =~ /^(endComp|endFile)$/) {
 		pop(@stack);
@@ -436,41 +463,50 @@ sub handle_preview_request_1
 	#
 	# Initial component trace
 	#
-	$trace = &$createTraceSub(@objects);
-	$pr->print("<h2>Component trace</h2>\n<pre>\n$trace</pre>\n");
+	if (@objects) {
+	    $trace = &$createTraceSub(@objects);
+	    $pr->print("<h2>Component trace</h2>\n<pre>\n$trace</pre>\n");
+	}
 
 	#
 	# Content
 	#
-	$pr->print("<h2>Content</h2>\n<pre>\n");
-	$content =~ s/\cM//g;
-	($content) = html_escape($content);
-	$content =~ s/\cAEVENT/\n\cAEVENT/g;
-	my ($obj);
-	while ($content =~ /(.*)/g) {
-	    my $line = $1;
-	    my $beginFlag = 0;
-	    if ($line =~ /\cAEVENT([0-9]+)\cA/) {
-		$eventnum = $1;
-		my $event = $compEvents[$eventnum];
-		$line =~ s/\cAEVENT[0-9]+\cA//;
-		if (defined($event)) {
-		    if ($event->{type} =~ /^end/ && $obj) {
-			$pr->print(sprintf("<font color=#%s>%5s &lt;&lt;</font>\n",$obj->{color},$obj->{label}));
-		    }
-		    $obj = $objects[$event->{objnum}];
-		    if ($event->{type} =~ /^start/) {
-			$pr->print("<a name=object$obj->{count}>");
-			$beginFlag = 1;
+	if ($content) {
+	    $pr->print("<h2>Content</h2>\n<pre>\n");
+	    $content =~ s/\cM//g;
+	    ($content) = html_escape($content);
+	    $content =~ s/\cAEVENT/\n\cAEVENT/g;
+	    my ($obj);
+	    while ($content =~ /(.*)/g) {
+		my $line = $1;
+		my $beginFlag = 0;
+		if ($line =~ /\cAEVENT([0-9]+)\cA/) {
+		    $eventnum = $1;
+		    my $event = $compEvents[$eventnum];
+		    $line =~ s/\cAEVENT[0-9]+\cA//;
+		    if (defined($event)) {
+			if ($event->{type} =~ /^end/ && $obj) {
+			    $pr->print(sprintf("<font color=#%s>%5s &lt;&lt;</font>\n",$obj->{color},$obj->{label}));
+			}
+			$obj = $objects[$event->{objnum}];
+			if ($event->{type} =~ /^start/) {
+			    $pr->print("<a name=object$obj->{count}>");
+			    $beginFlag = 1;
+			}
 		    }
 		}
+		next if ($line !~ /\S/ && !$obj);
+		my $depth = 0;
+		$pr->print(sprintf("<font color=#%s>%s%5s %s</font>%s%s\n",$obj->{color},(' ' x $depth),$obj->{label},($beginFlag ? '&gt;&gt;' : '  '),(' ' x (4-$depth)),$line));
 	    }
-	    next if ($line !~ /\S/ && !$obj);
-	    my $depth = 0;
-	    $pr->print(sprintf("<font color=#%s>%s%5s %s</font>%s%s\n",$obj->{color},(' ' x $depth),$obj->{label},($beginFlag ? '&gt;&gt;' : '  '),(' ' x (4-$depth)),$line));
+	    $pr->print("\n</pre>\n");
 	}
-	$pr->print("\n</pre>\n");
 
+	#
+	# Status
+	#
+	$pr->print("<h2>Status: $statuscode</h2>\n") if $statuscode && $statuscode != 200;
+	
         #
 	# Headers in
 	#
@@ -492,7 +528,7 @@ sub handle_preview_request_1
 	my %outHeaders = $r->headers_out();
 	my $gmtdate = gmtime;
 	$pr->print("<li type=circle><b>Date:</b>  $gmtdate\n");
-	my $serverType = $r->cgi_var('SERVER_SOFTWARE') || 'Apache/1.2 mod_perl/1.0'; 
+	my $serverType = $ENV{SERVER_SOFTWARE} || 'Apache/1.2 mod_perl/1.0'; 
 	$pr->print("<li type=circle><b>Server:</b>  $serverType\n");
 	while (($key,$value) = each(%outHeaders)) {
 	    $pr->print("<li type=disc><b>$key:</b>  $value\n");
